# See the file LICENSE for redistribution information
# 
# Copyright (c) 2002 -2004
# 	Sleepycat Software.  All rights reserved.
#
# $Id: xbench.tcl,v 1.5 2004/10/14 16:19:50 awood Exp $
#
# Standalone Tcl script to drive DB XML benchmark programs
# The XBench data needs to have been generated

# Error stream that (should!) always go to the console
set consoleerr stderr

# load up the containers
proc load_containers { { path_xbench_data "." } { path_dbenv "." } { path_executables "." } { size "tiny" } { type "wholedoc"} } {

    # help option (first argument is "?")
    if { $path_xbench_data == "?" } {
        puts "Usage: load_containers \[data=.\]  \[dbenv=.\] \[path=.\] \[size=tiny\] \[type=wholedoc\]\n"
        puts "Loads DB XML containers using XML data generated by the University of Waterloo XBench tools (http://db.uwaterloo.ca/~ddbms/projects/xbench/). Each invocation generates 4 containers - one each for DC-MD, DC-SD, TC-MD and TC-SD data. The size of these containers, and the storage type, can be specified.\n"
        puts "\"data\" is the path to the directory containing the XBench XML data (should sub directories \"TC-SD\" etc.).\n"
        puts "\"dbenv\" is the path to the DB XML environment directory in which the containers will be created.\n"
        puts "\"path\" is the path to the directory containing the DB XML executables used to load the container.\n"
        puts "\"size\" can be any of \"miniscule\" (container about 10-20K in size), \"tiny\" (about 1MB), \"small\" (about 10MB), \"medium\" (about 50MB), \"normal\" (about 100MB) or \"large\" (about 1GB).\n"
        puts "\"type\" can be either \"wholedoc\" or \"node\".\n"
        puts "The containers will have names like \"TC-SD-<size>-<type>.dbxml\".\n"
        puts "NOTE execution can take a long time for the larger containers!\n"

        return
    }

    set err 1

    # sort out container size
    set supported_sizes [list "miniscule" "tiny" "small" "medium" "normal" "large"]
    if { [lsearch -exact $supported_sizes $size] < 0 } {
        puts "Unsupported \"size\" argument: $size (use \"load_containers ?\")"
        return $err
    }

    # sort out storage type 
    set supported_types [list "wholedoc" "node"]
    if { [lsearch -exact $supported_types $type] < 0 } {
        puts "Unsupported \"type\" argument: $type (use \"load_containers ?\")"
        return $err
    }

    # verify the executable
    set program_name "load_container"
    set executable [file join $path_executables $program_name]
    if { [file exists $executable] != 1} {
        puts "Cannot find executable \"$executable\""
        return $err
    }

    # verify the XBench data directories
    set data_groups [list "DC-MD" "DC-SD" "TC-MD" "TC-SD"]
    foreach group $data_groups {
        set dir [file join $path_xbench_data $group $size]
        if {[file exists $dir] != 1} {
            puts "Cannot find the XBench data sub-directory \"$dir\""
            return $err
        }
    }

    puts "Loading \"$size\" containers of type \"$type\""

    # set up the db environment
    set dbenv [file join "." $path_dbenv]
    set action "Using"
    if { [file exists $dbenv] != 1 } {
        file mkdir $dbenv
        set action "Creating"
    }
    puts -nonewline "$action the DB XML environment \"$dbenv\"\n"

    # prepare program execution 
    set base_args {}
    set base_args [concat $base_args "-d $dbenv"]
    set base_args [concat $base_args "-s $type"]

    # iterate over data groups (one container / program run per group)
    set total_time_start [clock seconds]
    foreach group $data_groups {
        puts -nonewline "\n"
        set args $base_args

        # specify path to the XML files
        set path_to_files [file join $path_xbench_data $group $size]
        set args [concat $args "-p $path_to_files"]

        # XML files
        # build a filelist since number of arguments will be excessive
        set filelist_file FILELIST-$group
        set fh [open $filelist_file w]
        set ret [catch { glob $path_to_files/*.xml } result]
        if { $ret == 0 } {
            foreach filepath $result {
                puts $fh [file tail $filepath]
            }
        } else {
            puts "FAIL Could not find XML data files in \"\$path_to_files\""
            return $err
        }
        close $fh
        set args [concat $args "-f $filelist_file"]

        # container name - say if we are creating or replacing
        set container_name "$group-$size-$type.dbxml"
        set args [concat $args "-c $container_name"]

        set container_file [file join $dbenv $container_name]
        if { [file exists $container_file] != 1} {
            set action "Creating"
        } else {
            file delete -force $container_file
            set action "Replacing"
        }
        puts "\t$action the container \"$container_name\""

        # tidy up execution string
        regsub -all {\$} $args {\$} args
        set args [join $args]

        puts -nonewline "\tExecuting $executable ..."
        flush stdout
        set local_time_start [clock seconds]

        set ret [catch { eval exec $executable $args >> $program_name.out 2>> $program_name.err } res]

        if { $ret != 0 } {
            puts -nonewline "\n";
            puts "FAIL:\n\t$res\nCheck $program_name.err for more details."
        } else {
            puts -nonewline " OK"
            puts -nonewline "\n"
        }

        set local_time_end [clock seconds]
        set local_time_elapsed [expr $local_time_end - $local_time_start]
        puts -nonewline "\tTime: $local_time_elapsed seconds\n"

        file delete -force $filelist_file
    }

    puts -nonewline "\n"
    set total_time_end [clock seconds]
    set total_time_elapsed [expr $total_time_end - $total_time_start]
    puts -nonewline "Total time: $total_time_elapsed seconds\n"

    puts -nonewline "See $program_name.out and $program_name.err for runtime information.\n"

    set err 0
    return $err
}

proc query_container { { path_dbenv "." } { path_test_data "." } { path_executables "." } { size "tiny" } } {
    # help option (first argument is "?")
    if { $path_dbenv == "?" } {
        puts "Usage: query_container \[dbenv=.\] \[data=.\] \[path=.\] \[size=tiny\]\n"
        puts "Queries the containers that have been created by \"load_container\". using the XML Query Use Case workload from the University of Waterloo XBench tools (http://db.uwaterloo.ca/~ddbms/projects/xbench/)."
        puts "Each invocation queries 4 containers (containing DC-MD, DC-SD, TC-MD and TC-SD data) for each of NLS and WDS container types . The size of these containers can be specified.\n"
        puts "\"dbenv\" is the path to the DB XML environment directory.\n"
        puts "\"data\" is the path to the directory containing the test configuration data (contains e.g. \"queries-DC-SD.xml\").\n"
        puts "\"path\" is the path to the directory containing the DB XML executable used to run the queries.\n"
        puts "\"size\" can be any of \"miniscule\" (container about 10-20K in size), \"tiny\" (about 1MB), \"small\" (about 10MB), \"medium\" (about 50MB), \"normal\" (about 100MB) or \"large\" (about 1GB).\n"
        puts "NOTE The containers must exist (see \"load_container ?\")"
        puts "NOTE execution can take a long time for the larger containers!\n"
        return
    }

    set err 1

    # sort out container size
    set supported_sizes [list "miniscule" "tiny" "small" "medium" "normal" "large"]
    if { [lsearch -exact $supported_sizes $size] < 0 } {
        puts "Unsupported \"size\" argument: $size (use \"load_containers ?\")"
        return $err
    }

    # verify the dbenv
    if {[file exists $path_dbenv] != 1} {
        puts "Cannot find the DB XML environment directory \"$path_dbenv\""
        return $err
    }
    set ret [catch { glob $path_dbenv/*.dbxml } result]
    if { $ret != 0 } {
        puts "There are no *.dbxml files in the DB XML environment directory \"$path_dbenv\""
        return $err
    }

    # verify the test data directory
    set data_groups [list "DC-MD" "DC-SD" "TC-MD" "TC-SD"]
    foreach group $data_groups {
        set file "queries-$group.xml"
        set file [file join $path_test_data $file ]
        if {[file exists $file] != 1} {
            puts "Cannot find the config data file \"$file\""
            return $err
        }
    }

    # verify the executable
    set program_name "query_runner"
    set executable [file join $path_executables $program_name]
    if { [file exists $executable] != 1} {
        puts "Cannot find executable \"$executable\""
        return $err
    }

    puts "Querying \"$size\" containers..."

    # prepare program execution 
    set base_args {}

    set container_types [list "wholedoc" "node"]

    # iterate over data groups (each data group is run for NLS and WDS))
    set total_time_start [clock seconds]
    foreach group $data_groups {
        puts -nonewline "\n"
        puts "\tQuerying $group-$size.dbxml"

        set args $base_args

        # create a config file by wrapping appropriate database and container
        # instructions around the queries
        set query_file "queries-$group.xml"
        set query_file [file join $path_test_data $query_file ]

        set database $path_dbenv

        set config_file CONFIG-$group
        set fh [open $config_file w]

        puts $fh "<xqueryrunner database=\"$database\">"
        puts $fh "<group name =\"$group-$size\">"
        puts $fh "<containers>"

        foreach container_type $container_types {
            puts $fh "<container>$group-$size-$container_type.dbxml</container>"
        }

        puts $fh "</containers>"
        puts $fh "<queries>"
        puts $fh [readFileUTF8 $query_file]
        puts $fh "</queries>"
        puts $fh "</group>"
        puts $fh "</xqueryrunner>"

        close $fh
        set args [concat $args "-c $config_file"]

        # redirect output to an intermediate file
        #  2004-10-06 ARW leave the user to do something with the results, e.g.
        #    stylesheet - so do _not_ delete this file!
        set out_file "benchmark-query-results-$group-$size.xml"
        set args [concat $args "-f $out_file"]

        # tidy up execution string
        regsub -all {\$} $args {\$} args
        set args [join $args]

        puts -nonewline "\tExecuting $executable ..."
        flush stdout
        set local_time_start [clock seconds]

        set ret [catch { eval exec $executable $args >> $program_name.out 2>> $program_name.err } res]

        if { $ret != 0 } {
            puts -nonewline "\n";
            puts "FAIL:\n\t$res\nCheck $program_name.err for more details."
        } else {
            puts -nonewline " OK"
            puts -nonewline "\n"
            puts -nonewline "\tResults written to \"$out_file\"\n"
        }

        set local_time_end [clock seconds]
        set local_time_elapsed [expr $local_time_end - $local_time_start]
        puts -nonewline "\tTime: $local_time_elapsed seconds\n"

        file delete -force $config_file
        #file delete -force $out_file
    }

    puts -nonewline "\n"
    set total_time_end [clock seconds]
    set total_time_elapsed [expr $total_time_end - $total_time_start]
    puts -nonewline "Total time: $total_time_elapsed seconds\n"
    puts -nonewline "See $program_name.out and $program_name.err for runtime information.\n"

    puts -nonewline "\n"
    puts -nonewline "TODO analyse the output (\"$program_name-results-*.xml\"), e.g. apply a stylesheet to extract error and timing information.\n"
    puts -nonewline "There should be no \"<error>\" elements, and the query count should be as follows:\n"
    puts -nonewline "\tDC-MD 15\n"
    puts -nonewline "\tDC-SD 16\n"
    puts -nonewline "\tTC-MD 19\n"
    puts -nonewline "\tTC-SD 17\n"
    puts -nonewline "\n"

    set err 0
    return $err
}

proc help {} { 
    puts "Methods for performance and benchmark testing."
    puts "\n"
    puts "load_containers\n\tLoads up DB XML containers with data generated by XBench.\n\tType \"load_containers ?\" for further information."
    puts "\n"
    puts "query_container\n\tRuns the Use Case queries on the containers.\n\tType \"query_container ?\" for further information."
    puts "\n"
    puts "The XBench data needs to have been generated - visit \"http://db.uwaterloo.ca/~ddbms/projects/xbench/\"."
}

proc readFileUTF8 { filename } {
	set fd [open $filename r]
	fconfigure $fd -encoding utf-8
	set data [read $fd]
	close $fd
	return $data
}
