/*
 * Copyright (c) 2001, DecisionSoft Limited All rights reserved.
 * Please see LICENSE.TXT for more information.
 */

#include "../config/pathan_config.h"
#include <pathan/internal/XPath2MemoryManagerImpl.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLException.hpp>
#include <xercesc/util/RuntimeException.hpp>
#include <xercesc/util/HashPtr.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/dom/DOMDocument.hpp>

#include <pathan/internal/StringPool.hpp>
#include <pathan/PathanException.hpp>
#include <pathan/internal/context/XPath2ContextImpl.hpp>
#include <pathan/internal/DOMutils/XStr.hpp>
#include <pathan/internal/collations/CollationImpl.hpp>
#include <pathan/internal/dom-extensions/PathanNSResolverImpl.hpp>
#include "../simpleVariables/VarStoreImpl.hpp"
#include "../simpleVariables/VarTypeStoreImpl.hpp"
#include "../functionAPI/FunctionLookupImpl.hpp"
#include "../items/impl/ATDecimalOrDerivedImpl.hpp"
#include <pathan/functions/FunctionConstructor.hpp> // for datatype-uri

#include <iostream>

XPath2MemoryManagerImpl::XPath2MemoryManagerImpl()
{
  initialise();
}

XPath2MemoryManagerImpl::~XPath2MemoryManagerImpl() 
{
  releaseAll();
}

void XPath2MemoryManagerImpl::reset()
{
  releaseAll();
  initialise();
}

void XPath2MemoryManagerImpl::initialise()
{
#ifdef WIN32
	fHeap=HeapCreate(HEAP_NO_SERIALIZE,128*1024,0);
#else
  fCurrentBlock = 0;
#endif
  fStringPool = new (this) StringPool(this);
  fIntegerPool = new (this) XERCES_CPP_NAMESPACE_QUALIFIER RefHashTableOf<ATDecimalOrDerived>(53,true, new (this) XERCES_CPP_NAMESPACE_QUALIFIER HashPtr(),this);
}

void XPath2MemoryManagerImpl::releaseAll()
{
  fIntegerPool->cleanup();
  //  Delete the heap for this document.  This uncerimoniously yanks the storage
  //      out from under all of the nodes in the document.  Destructors are NOT called.
#ifdef WIN32
	HeapDestroy(fHeap);
#else
  // Run backwards through the linked list, deleting the blocks of memory
  while(fCurrentBlock) {
    MemList *prev = fCurrentBlock->prev;
    delete [] ((char *)fCurrentBlock);
    fCurrentBlock = prev;
  }
#endif
}

void *XPath2MemoryManagerImpl::allocate(size_t amount)
{
#ifdef WIN32
	return HeapAlloc(fHeap,HEAP_NO_SERIALIZE|HEAP_ZERO_MEMORY,amount);
#else
  char *newBlock = 0;
  try {
    newBlock = new char[amount + sizeof(MemList)];
  }
  catch (...) {}
  if(!newBlock) {
    //ThrowXML(RuntimeException, XMLExcepts::Out_Of_Memory);
    throw PathanException(PathanException::RUNTIME_ERR, X("XPath2MemoryManagerImpl::allocate(): Out of memory"));
  }

  MemList *newMemList = (MemList *)newBlock;
  newMemList->prev = fCurrentBlock;
  newMemList->next = 0;

  if(fCurrentBlock) {
    fCurrentBlock->next = newMemList;
  }
  fCurrentBlock = newMemList;

  return (void *)(newBlock + sizeof(MemList));
#endif
}

void XPath2MemoryManagerImpl::deallocate(void* p)
{
#ifdef WIN32
	HeapFree(fHeap,HEAP_NO_SERIALIZE,p);
#else
  if(p) {
    char *oldBlock = ((char *)p) - sizeof(MemList);
    MemList *oldMemList = (MemList *)oldBlock;

    // Decouple the old block from the list
    if(oldMemList->prev) {
      oldMemList->prev->next = oldMemList->next;
    }
    if(oldMemList->next) {
      oldMemList->next->prev = oldMemList->prev;
    }
    else {
      // If the old block doesn't have a next, it must currently be fCurrentBlock
      fCurrentBlock = oldMemList->prev;
    }

    delete [] oldBlock;
  }
#endif
}

const XMLCh *XPath2MemoryManagerImpl::getPooledString(const XMLCh *src)
{
  return fStringPool->getPooledString(src);
}

const XMLCh *XPath2MemoryManagerImpl::getPooledString(const char *src)
{
  return fStringPool->getPooledString(src);
}

void XPath2MemoryManagerImpl::dumpStatistics()
{
  std::cout << "Memory statistics:" << std::endl;

  std::cout << "\tStrings in pool: " << fStringPool->getCount() << std::endl;

#ifndef WIN32
  MemList* pBlockCursor = fCurrentBlock;
  unsigned int nBlockCount=0;
  while(pBlockCursor) {
    ++nBlockCount;
    pBlockCursor = pBlockCursor->prev;
  }
  std::cout << "\tAllocated objects: " << nBlockCount << std::endl;
#endif
}

/** create a collation */
Collation* XPath2MemoryManagerImpl::createCollation(CollationHelper* helper) {
  return new (this) CollationImpl(this,helper);
}

/** create a resolver */
PathanNSResolver* XPath2MemoryManagerImpl::createNSResolver(XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *resolverNode) {
  return new (this) PathanNSResolverImpl(this, resolverNode);
}

/** create a store for variables */
VariableStore* XPath2MemoryManagerImpl::createVariableStore() {
  return new (this) VarStoreImpl(this);
} 

/** create a store for variables */
VariableTypeStore* XPath2MemoryManagerImpl::createVariableTypeStore() {
  return new (this) VarTypeStoreImpl(this);
} 

/** create a function table */
FunctionLookup* XPath2MemoryManagerImpl::createFunctionTable() {
  return new (this) FunctionLookupImpl(this);
}

/** create a ATDecimalOrDerived for the given integer */
ATDecimalOrDerived* XPath2MemoryManagerImpl::createInteger(int value) {
  if(fIntegerPool->containsKey((const void*)value))
      return fIntegerPool->get((const void*)value);
  ATDecimalOrDerivedImpl* itemValue=new ATDecimalOrDerivedImpl(value);
  itemValue->incrementRefCount();
  fIntegerPool->put((void*)value, itemValue);
  return itemValue;
}
